import code
import random
import string
import sys
import uuid

import click
import httpx
from praw import Reddit
from rich import pretty
from rich.status import Status
from rich.traceback import install
from shiv.bootstrap import current_zipfile

import src
from src.helpers import flip_char, load_config, write_config
from src.joplin import process_joplin_posts, get_folders, BASE_URL
from src.art import BANNERS


@click.group(
    context_settings=dict(help_option_names=["-h", "--help", "--halp"]),
    invoke_without_command=True,
)
@click.pass_context
@click.version_option(version=src.__version__, prog_name="utils")
@click.option(
    "--update",
    is_flag=True,
    help="Check Gitea for a new version and auto-update.",
)
def main(ctx, update):
    """Launch a utility or drop into a command line REPL if no command is given."""
    if update:
        update_from_gitea()
        sys.exit()
    elif ctx.invoked_subcommand is None:
        banner = random.choice(BANNERS)

        pretty.install()  # type: ignore
        install()  # traceback handler

        code.interact(local=globals(), banner=banner)
        sys.exit()


@main.command()
def uuid4():
    """Generate a random UUID4."""
    click.echo(uuid.uuid4())


@main.command()
def joplin():
    """Search Joplin for notes titled 'convert' and process them."""
    process_joplin_posts()


@main.command()
def objectid():
    """Generate a random ObjectID."""
    new_id = ""
    for _ in range(24):
        new_id += random.choice(string.ascii_lowercase[:6] + string.digits)
    click.echo(new_id)


@main.command()
@click.argument("dice")
def roll(dice: str):
    """Roll some dice. Format: utils roll 3d8"""
    if "d" not in dice:
        click.echo("Missing part of the call. Example: 1d10")
        return
    if len(dice.split("d")) != 2:
        click.echo("Error parsing dice. Example: 2d6")
        return
    num, sides = dice.split("d")
    try:
        num = int(num)
        sides = int(sides)
    except ValueError:
        click.echo("Need numbers for the dice. Example: 30d4")
        return
    if num < 1 or sides < 1:
        click.echo("Dude. Example: 2d8")
        return
    values: list[int] = []
    for die in range(num):
        values.append(random.randint(1, sides))

    if num == 1:
        click.echo(f"You rolled a {values[0]}.")
    else:
        click.echo(f"You rolled: {'+'.join([str(item) for item in values])}")
        click.echo(f"Total: {sum(values)}")


@main.command()
@click.argument("words", nargs=-1)
def beautify(words: list[str]):
    """
    MAkE YoUr mEsSaGe bEaUtIfUl!!!1!!

    WORDS is either a single string surrounded by double quotes or multiple bare words,
    e.g. `utils beautify "one two three"` or `utils beautify one two three`.
    """
    message = " ".join(words)
    new_beautiful_string = []

    for num, letter in enumerate(message):
        if num % 2:
            letter = flip_char(letter)
        new_beautiful_string.append(letter)

    click.echo("".join(new_beautiful_string))


@main.command()
def get_saved_from_reddit() -> None:
    """Get saved posts from reddit."""

    def _process(item):
        fullname = item.fullname
        if fullname.startswith("t3"):
            # we got a post
            title = f"{item.subreddit.display_name} - {item.title}"
            body = item.selftext if item.selftext else item.url
        elif fullname.startswith("t1"):
            # comment time
            try:
                author_name = item.author.name
            except AttributeError:
                author_name = "[deleted]"
            title = (
                f"{item.submission.subreddit.display_name} - {item.submission.title}"
            )
            body = f"Comment from {author_name}:\n\n{item.body}"
        else:
            click.echo(f"Not sure how to process https://reddit.com{item.permalink}")
            return

        if item.over_18:
            title = "🔴 " + title
        body = "https://reddit.com" + item.permalink + "\n\n" + body

        joplin = BASE_URL + cfg["JOPLIN_PORT"]
        notes_url = joplin + f"/notes/?token={cfg.get('JOPLIN_TOKEN')}"
        click.echo(f"Processing {title}...")
        httpx.post(
            notes_url,
            json={
                "title": title,
                "body": body,
                "parent_id": cfg.get("joplin_saved_posts_folder_id"),
            },
        )
        item.unsave()

    cfg = load_config()
    # because 2fa is enabled, we have to get the code first
    twofa_code = click.prompt("What's the current 2fa code?", type=str)
    if not cfg.get("REDDIT_CLIENT_ID"):
        cfg["REDDIT_CLIENT_ID"] = ""
    if not cfg.get("REDDIT_CLIENT_SECRET"):
        cfg["REDDIT_CLIENT_SECRET"] = ""
    if not cfg.get("REDDIT_PASSWORD"):
        cfg["REDDIT_PASSWORD"] = ""
    if not cfg.get("REDDIT_USERNAME"):
        cfg["REDDIT_USERNAME"] = ""
    if not cfg.get("joplin_saved_posts_folder_id"):
        cfg["joplin_saved_posts_folder_id"] = ""

    write_config(cfg)

    username = cfg.get("REDDIT_USERNAME", "")

    r = Reddit(
        client_id=cfg.get("REDDIT_CLIENT_ID", ""),
        client_secret=cfg.get("REDDIT_CLIENT_SECRET", ""),
        username=username,
        password=f"{cfg.get('REDDIT_PASSWORD', '')}:{twofa_code}",
        user_agent=f"getter of saved posts, u/{username}",
    )

    try:
        r.user.me()
    except Exception as e:
        click.echo(f"Cannot connect to reddit: {e}")
        return

    if not cfg.get("joplin_saved_posts_folder_id"):
        folders = get_folders()
        click.echo("Pick the folder that I should write your saved posts to:")
        for count, option in enumerate(folders["items"]):
            click.echo(f"{count} - {option['title']}")
        folder_position = click.prompt("Number of folder?", type=int)
        folder_name = folders["items"][folder_position]["title"]
        folder_id = folders["items"][folder_position]["id"]

        click.echo(f"Got it, will write to {folder_name}, ID {folder_id}.")
        cfg["joplin_saved_posts_folder_id"] = folder_id
        write_config(cfg)

    for _ in range(10):
        click.echo("Getting new posts...")
        for item in r.user.me().saved(limit=None):
            _process(item)


def update_from_gitea():
    """Get the newest release from Gitea and install it."""
    status = Status("Checking for new release...")
    status.start()

    response = httpx.get(
        "https://git.joekaufeld.com/api/v1/repos/jkaufeld/utils/releases/latest"
    )
    if response.status_code != 200:
        status.stop()
        click.echo(
            f"Something went wrong when talking to Gitea; got a"
            f" {response.status_code} with the following content:\n"
            f"{response.content}"
        )
        return
    status.update("Checking for new release...")
    release_data = response.json()
    if release_data["tag_name"] == src.__version__:
        status.stop()
        click.echo("Server version is the same as current version; nothing to update.")
        return
    status.update("Updating...")

    url = release_data["assets"][0]["browser_download_url"]
    with current_zipfile() as archive:
        with open(archive.filename, "wb") as f, httpx.stream(
            "GET", url, follow_redirects=True
        ) as r:
            for line in r.iter_bytes():
                f.write(line)

    status.stop()
    click.echo(f"Updated to {release_data['tag_name']}! 🎉")


if __name__ == "__main__":
    main()
